/* CS-PSeq-Gen specific part
 * Management of clusters of co-subtituting sites
 *
 * by P. Tuffery (1997-2000)
 * INSERM U436, Universite Paris 7, France
 */


#ifndef _CORRTYPES_H_
#define _CORRTYPES_H_

enum { SIM, COR, BOT, numMutation };

#if 0
typedef struct scorrRec
{
  double UsrMatrix[16];         /* copy SpecificMatrix in SpecificMatrix2 before (necessary) addition by row to   */
  double SpeMat2[16];           /* write original values of SpecificMatrix (or vector) in outfile 'resumCorr.dat' */
  double UsrVector[4];
  double SpeVec2[16];
  int nCorr;
  int *corr;
  int proto;                    /* Prototype for cluster of correlated sites */
  int nMut;
  double tFac;                  /* Tansmission of mutations through each cluster element */
  char MutAllowed;
  int  conv;                    /* Conversion towards PDB number from 1 */
  double sMut;                  /* Non correlated part of mutation for site */
} DtCorrRec;
#endif

#define DcCSDISABLED 0
#define DcCSENABLED  1

#define DcCONSTANT 0
#define DcEVOLVING 1
#define DcDRIVER   2
#define DcMEMBER   4


#define DcMAXCOMUT     50    /* Max number of so-substitutions par site pair */
#define DcMAXGLOBCOMUT 50000 /* Max number of so-substitutions per tree */

/* ===========================================================
 * This structure to pass parameters to define clusters
 * of co-substituting sites
 * ===========================================================
 */
typedef struct sclusrec {
  int    rndDefine;   /* Do we define clusters randomly ?  */
  int    nCls;        /* numbre of clusters to define      */
  int    minSze;      /* Minimal size of clusters          */
  int    maxSze;      /* Maximal size of clusters          */
  double minRate;     /* min heterogeneity rate for member */
  double maxRate;     /* max heterogeneity rate for member */
  double maxDiffRate; /* max diff rate to driver           */
  double tFac;        /* transmission factor               */
} DtCSClsDef;

typedef struct smutprf
{
  int tpe;         /* class of site (DcCONSTANT,DcEVOLVING,DcSEED,DcMEMBER) */
  int nCss;        /* Number of co-subsituting sites                   */
  int lnMut;       /* To trace number of time cluster has undergone CM */
  int nMut;        /* To trace number of time cluster has undergone CM */
  int *cs;
  int drvr;        /* Index of the driver of the cluster               */
  double tFac;     /* Tansmission of substitution to cluster element   */
  int    conv;     /* Conversion towards PDB number from 1             */
  double rFac;     /* Non correlated part of mutation for site         */
} DtSiteCSRec;


/* ===========================================================
 * This structure to pass parameters during simulations.
 * ===========================================================
 */
typedef struct scsrec {
  int          sze;     /* number of sites (numBases)      */
  int          nES;     /* number of truly evolving sites  */
  int          tFac;    /* Default tFac                    */
  int          nCls;    /* number of clusters              */
  int          nClsEle; /* Number of sites in clusters     */
  int          active;  /* Evolution uses CS               */
  int        **rfCsTab; /* count of co-subsitutions for reference set    */
  int          nRfCs;   /* Overall number of co-subsitutions for ref set */
  int        **csTab;   /* count of co-subsitutions for simulated set    */
  int          nCs;     /* Overall number of co-subsitutions for sim set */
  int        **rsTab;   /* Combination of rfCsTab and csTab              */
  int       ***laws;    /* Law per site pair                             */
  int        **trhld;   /* Threshold per site pair                       */
  double     **toss; 
  int          gLaw[DcMAXGLOBCOMUT];
  DtSiteCSRec *p;       /* Pointer towards each site infos               */
} DtCSRec;



#endif /* _CORRELAT_H_ */






