/* 
 * Some random utilities
 * 
 * by P. Tuffery (1997-2000) 
 * INSERM U436, Universite Paris 7, France
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/times.h>

long int gSeed = -1;

/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
void argstrRndPrmtrsUsage(int argc,char *argv[])
{
/*    delimiter(stderr); */
  fprintf(stderr,"Pseudo Random number generator initialisation\n");
  fprintf(stderr," -seed l     : seed for drand48 pseudo-random number generator\n");
}

/* =====================================================================
 * Parse command line arguments
 * =====================================================================
 */
void parseArgstrRndPrmtrs(int argc,char *argv[])
{
  int i;
  for (i=0;i<argc;i++) {
    if (!strcmp(argv[i],"-seed")) {     /* drand48 seed */
      gSeed = atol(argv[i+1]);  
      ++i;
    }
  }
}

/* =====================================================================
 * Global random parameters information
 * =====================================================================
 */
void rndSumary(FILE *f)
{
  fprintf(f," Peudo-Random Generator Seed : %ld\n", gSeed);
}

/* =====================================================================
 * Initialize pseudo-random generator.
 * returns the seed value.
 * =====================================================================
 */
void initRnd(long int *seed)
{
  static isInit = 0;

  if (isInit) return;
  if ((int) *seed == -1) {
    *seed = (long int) time(0);
    srand48(*seed);
  }
  else srand48(*seed);

  isInit = 1;
}

/* Returns a random integer from 0 to n (excluded) ===================== */
int irand(int n)
{
  extern double drand48();
  double rn;
  float rnf;
  
  rn   = drand48();		/* suppose drand48 does not accept 1.0	*/
  rnf  = (float) rn * n;
  
  return (int) rnf;
}

/* Returns a random double from x (included) to y (excluded) ===================== */
double drandxy(double x, double y)
{
  extern double drand48();
  double rn;
  
  rn   = drand48();		/* supposes drand48 does not accept 1.0	*/
  rn   = rn * (y-x) + x;
  if (rn >= y) rn = y;
  return rn;
}

/* Returns a random integer from 0 to n (excluded) pos excluded ============= */
int irandnotn(int n,int pos)
{
  extern double drand48();
  
  double rn;
  float rnf;
  
  do {
  rn   =  drand48();		/* suppose drand48 does not accept 1.0	*/
  rnf  = (float) rn * n;
  } while ((int) rnf == pos);
  
  return (int) rnf;
}
  
