#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "global.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "random.h"
#include "CSFuncs.h"
#include "Rates.h"
#include "TreeUtils.h"  

extern int DEBUG;

/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
void argstrRefSeqPrmtrsUsage()
{
  fprintf(stderr,"Reference set parameters\n");
  fprintf(stderr,"  -iRefNames <fname>     : ref labels (PAUP) (tips and nodes)\n");
  fprintf(stderr,"  -iPAMLRefNames <fname> : ref labels (PAML) (tips and nodes)\n");
  fprintf(stderr,"  -iRefSeq <fname>       : ref sequences (PAUP) (tips and nodes)\n");
  fprintf(stderr,"  -iPAMLRefSeq <fname>   : ref sequences (PAML) (tips and nodes)\n");
  fprintf(stderr,"  -refSeqBLen            : branch lengths from number of susbtitutions along reference sequences\n\n");
}

/* =====================================================================
 * Parse CS command line arguments
 * =====================================================================
 */
void parseArgstrRefSeqPrmtrs(int argc,char *argv[])
{
  extern char iRefSeqFileName[BUFSIZ];
  extern char iPAMLRefSeqFileName[BUFSIZ];
  extern char iRefNamesFileName[BUFSIZ];
  extern char iPAMLRefNamesFileName[BUFSIZ];
  extern int  gUseRefBLen;
  int i;

  for (i=0;i<argc;i++) {

    if (!strcmp(argv[i],"-iRefNames")) {
      strcpy(iRefNamesFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iPAMLRefNames")) {
      strcpy(iPAMLRefNamesFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iRefSeq")) {
      strcpy(iRefSeqFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-iPAMLRefSeq")) {
      strcpy(iPAMLRefSeqFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-refSeqBLen")) {
      gUseRefBLen = 1;
      continue;
    }
  }
}

/* =========================================================
 * Check if new ref name is consistent with previously
 * determined ones
 * =========================================================
 */
static int checkNewRefName(TTree *tree, int i, char *node)
{
  if (tree->IntermediateNode[i]->refName == NULL) {
    tree->IntermediateNode[i]->refName = malloc((strlen(node) + 1) * sizeof(char));
    strcpy(tree->IntermediateNode[i]->refName,node);
/*     fprintf(stderr,"Found node_%s ",node); */
    if ((tree->IntermediateNode[i]->branch1 != NULL) && (tree->IntermediateNode[i]->branch1->refName != NULL)) {
/*       fprintf(stderr," --> %s",tree->IntermediateNode[i]->branch1->refName); */
    }
    if ((tree->IntermediateNode[i]->branch2 != NULL) && (tree->IntermediateNode[i]->branch2->refName != NULL)) {
/*       fprintf(stderr," --> %s",tree->IntermediateNode[i]->branch2->refName); */
    }
/*     fprintf(stderr,"\n"); */
    return 1;
  } else if (strcmp(tree->IntermediateNode[i]->refName,node)) {
    fprintf(stderr,"Error: Inconsistency while reading Ref Names (%s / %s)\n",tree->IntermediateNode[i]->refName,node);
    fprintf(stderr,"Was node_%s ",tree->IntermediateNode[i]->refName);
    if ((tree->IntermediateNode[i]->branch1 != NULL) && (tree->IntermediateNode[i]->branch1->refName != NULL)) {
      fprintf(stderr," --> %s",tree->IntermediateNode[i]->branch1->refName);
    }
    if ((tree->IntermediateNode[i]->branch2 != NULL) && (tree->IntermediateNode[i]->branch2->refName != NULL)) {
      fprintf(stderr," --> %s",tree->IntermediateNode[i]->branch2->refName);
    }
    fprintf(stderr,"\n");
    
    exit(0);
  }
  return 0;
}

/* =========================================================
 * Load the names assoicated with the taxa sequences
 * AND THE NODES
 * =========================================================
 */
void inputRefNames(TTree *tree, char *fname, FILE *flis)    /* PT 98 */
{
    int i, j,k,l,b=0;
    char *P;
    FILE *fv;
    int numtipnode;
    char line[BUFSIZ];
    char node[100], cmt[20], bnode[100], *pNode, *pBNode;
    int newRef;

    if (fname[0] == '\0') {
      fprintf(stderr,"inputRefNames: Null file name\n");
      return;
    }
    fv=fopen(fname, "r");
    if (fv == NULL) {
      fprintf(stderr, "Sorry: could not open file %s for node reference names assignment\n",fname);
      exit(0);
    }
    fprintf(stderr, "inputRefNames File: %s\n",fname);

    if (tree->rooted) {
#if 0
      tree->root->refName = malloc((strlen("root") + 1) * sizeof(char));
      strcpy(tree->root->refName,"root");
#endif
      tree->mask = malloc(sizeof(int) * (numBases+1));
      tree->root->sequence=CAllocMem(numBases+1, 
				     "sequences", "inputRefNames", 0);
      tree->root->refSeq  =CAllocMem(numBases+1, 
				     "refSeq", "inputRefNames", 0);
      for (i=0; i<numBases; i++) tree->mask[i] = 1;
    }
    for (i=0; i<tree->numTips; i++) {
      /* On copie le nom des sequences taxons comme refName */
      if (tree->tips[i]->refName == NULL) {
	tree->tips[i]->refName = malloc((strlen(tree->names[i]) + 1) * sizeof(char));
	strcpy(tree->tips[i]->refName,tree->names[i]);
      }
    }

    /* Pour les noeuds, on lit le fichier  de description de l'arbre */
    do {
      newRef = 0;
      rewind(fv);
      while (fgets(line, BUFSIZ, fv) != NULL)  {

	/* Lecture d'une ligne de la forme: node_XX --> node_YY (autre noeud) ou YY (taxon)
/* 	fprintf(stderr," %s \n",line); */
	if ((strstr(line,"node_") != NULL) && (strstr(line,"-->") != NULL)) {
	  sscanf(line,"%s %s %s",node,cmt,bnode);
	  pNode = node + 5;
	  pBNode = (strstr(bnode,"node_") != NULL) ? bnode + 5 : bnode;
/*  	  fprintf(stderr," %s --> %s\n",pNode,pBNode); */
	}

	/* Recherche de qui est XX et qui est YY */
	for (i=0; i<tree->numInteNode; i++) {
/* 	  fprintf(stderr,"Node %d\n",i); */
	  if (tree->IntermediateNode[i]->tipNo!=-1) continue;

	  if ((tree->IntermediateNode[i]->branch1 != NULL) && (tree->IntermediateNode[i]->branch1->tipNo != -1) &&
	      (!strcmp(tree->names[tree->IntermediateNode[i]->branch1->tipNo],bnode))) {
	    newRef += checkNewRefName(tree,i,pNode);
	  } else if ((tree->IntermediateNode[i]->branch2 != NULL) && (tree->IntermediateNode[i]->branch2->tipNo != -1) &&
	      (!strcmp(tree->names[tree->IntermediateNode[i]->branch2->tipNo],bnode))) {
	    newRef += checkNewRefName(tree,i,pNode);
	  } else if ((tree->IntermediateNode[i]->branch1 != NULL) && (tree->IntermediateNode[i]->branch1->tipNo == -1)) {
	    if ((tree->IntermediateNode[i]->branch1->refName != NULL) && 
		(!strcmp(tree->IntermediateNode[i]->branch1->refName,pBNode))) {
	      newRef += checkNewRefName(tree,i,pNode);
	    }
	  } else if ((tree->IntermediateNode[i]->branch2 != NULL) && (tree->IntermediateNode[i]->branch2->tipNo == -1)) {
	    if ((tree->IntermediateNode[i]->branch2->refName != NULL) && 
		(!strcmp(tree->IntermediateNode[i]->branch2->refName,pBNode))) {
	      newRef += checkNewRefName(tree,i,pNode);
	    }
	  }
	}
      }
    } while (newRef);

    /* Une petite verification ... */
#if 0
    for (i=0; i<tree->numInteNode; i++) {
      if (tree->IntermediateNode[i]->tipNo!=-1) continue;
      fprintf(stderr,"node_%s --> (%s , %s)\n",
	      tree->IntermediateNode[i]->refName,
	      tree->IntermediateNode[i]->branch1->refName,
	      tree->IntermediateNode[i]->branch2->refName);
    }
#endif
    if (flis != NULL) {
      fprintf(flis,"\nTree description using refNames:\n");
      DescribeTree(flis,tree,tree->root);
    }
}

/* =========================================================
 * Load the names assoicated with the taxa sequences
 * AND THE NODES
 * =========================================================
 */
void inputPAMLRefNames(TTree *tree, char *fname, FILE *flis)    /* PT 98 */
{
  int i, j,k,l,b=0, curTok, nTok;
  char *P, *pC;
  FILE *fv;
  int numtipnode;
  char line[BUFSIZ];
  char node[100], cmt[20], bnode[100], pNode[20], pBNode[20], *aToken;
  int newRef;
  char tipNames[MAX_TIPS][30];
  char *tok[MAX_TIPS*3];

  if (fname[0] == '\0') {
    fprintf(stderr,"inputPAMLRefNames: Null file name\n");
    return;
  }
    
  fv=fopen(fname, "r");
  if (fv == NULL) {
    fprintf(stderr, "inputPAMLRefNames: Sorry: could not open file %s for ref names assignment\n",fname);
    exit(0);
  }
  if (DEBUG) fprintf(stderr, "inputPAMLRefNames File: %s\n",fname);

  if (tree->rooted) {
#if 0
    tree->root->refName = malloc((strlen("root") + 1) * sizeof(char));
    strcpy(tree->root->refName,"root");
#endif
    tree->mask = malloc(sizeof(int) * (numBases+1));
    tree->root->sequence=CAllocMem(numBases+1, 
				   "sequences", "inputRefNames", 0);
    tree->root->refSeq  =CAllocMem(numBases+1, 
				   "refSeq", "inputRefNames", 0);
    for (i=0; i<numBases; i++) tree->mask[i] = 1;
  }

  /* For tips, refName is tip name */
  for (i=0; i<tree->numTips; i++) {
    /* On copie le nom des sequences taxons comme refName */
    if (tree->tips[i]->refName == NULL) {
      tree->tips[i]->refName = malloc((strlen(tree->names[i]) + 1) * sizeof(char));
      strcpy(tree->tips[i]->refName,tree->names[i]);
    }
  }

  /* Pour les noeuds, on lit d'abord les noms dans l'ordre PAML */
  for (i=0; i<tree->numTips; i++) {
    if (fgets(line, BUFSIZ, fv) == NULL) {
      fprintf(stderr, "inputPAMLRefNames: Sorry: file %s seems badly formatted\n",fname);
      exit(0);
    }
    sscanf(line,"%s",tipNames[i]);
  }
  
  /* Pour les noeuds, on lit le fichier  de description de l'arbre */
  do {
    newRef = 0;
    rewind(fv);

    /* 1 blank line before branch description ----------------- */
    for (i=0; i<=tree->numTips; i++)  if (fgets(line, BUFSIZ, fv) == NULL);
    /* The line for branch description ------------------------ */
    if (fgets(line, BUFSIZ, fv) == NULL) {
      fprintf(stderr, "inputPAMLRefNames: Sorry: file %s seem empty\n",fname);
      exit(0);
    }

    if ((pC = strchr(line,'\n')) != NULL) *pC = '\000';
    /* Parsing the line for one branch description ------------ */
    nTok = 0;
    tok[0] = strtok(line," ");
    while ((tok[++nTok] = strtok(NULL," ")) != NULL);

    for (curTok=0; curTok< nTok; curTok++) {

      aToken = tok[curTok];
      while ((pC = strchr(aToken,'.')) != NULL) *pC = ' ';
      sscanf(aToken,"%s%s",node, bnode);

      /* number to name conversion for tips */
      if (atoi(node) <= tree->numTips) strcpy(node,tipNames[atoi(node) - 1]);
      if (atoi(bnode) <= tree->numTips) strcpy(bnode,tipNames[atoi(bnode) - 1]);

/*        fprintf(stderr,"%s: %s -> %s\n",aToken,node,bnode); */
      sprintf(pNode,"#%s",node);
      sprintf(pBNode,"#%s",bnode);

      /* Recherche de qui est XX et qui est YY */
      for (i=0; i<tree->numInteNode; i++) {
	/* 	  fprintf(stderr,"Node %d\n",i); */
	if (tree->IntermediateNode[i]->tipNo!=-1) continue; /* Skip tips (already done ...) */
	
	/* branch1 leads to a tip ------------------------------ */
	if ((tree->IntermediateNode[i]->branch1 != NULL) && (tree->IntermediateNode[i]->branch1->tipNo != -1) &&
	    (!strcmp(tree->names[tree->IntermediateNode[i]->branch1->tipNo],bnode))) {
	  newRef += checkNewRefName(tree,i,pNode);
	/* branch2 leads to a tip ------------------------------ */
	} else if ((tree->IntermediateNode[i]->branch2 != NULL) && (tree->IntermediateNode[i]->branch2->tipNo != -1) &&
		   (!strcmp(tree->names[tree->IntermediateNode[i]->branch2->tipNo],bnode))) {
	  newRef += checkNewRefName(tree,i,pNode);
	/* branch1 leads to a node ----------------------------- */
	} else if ((tree->IntermediateNode[i]->branch1 != NULL) && (tree->IntermediateNode[i]->branch1->tipNo == -1)) {
	  if ((tree->IntermediateNode[i]->branch1->refName != NULL) && 
	      (!strcmp(tree->IntermediateNode[i]->branch1->refName,pBNode))) {
	    newRef += checkNewRefName(tree,i,pNode);
	  }
	/* branch2 leads to a node ----------------------------- */
	} else if ((tree->IntermediateNode[i]->branch2 != NULL) && (tree->IntermediateNode[i]->branch2->tipNo == -1)) {
	  if ((tree->IntermediateNode[i]->branch2->refName != NULL) && 
	      (!strcmp(tree->IntermediateNode[i]->branch2->refName,pBNode))) {
	    newRef += checkNewRefName(tree,i,pNode);
	  }
	}
      }
    }
  } while (newRef);

  /* Une petite verification ... */
#if 0
  for (i=0; i<tree->numInteNode; i++) {
    if (tree->IntermediateNode[i]->tipNo!=-1) continue;
    fprintf(stderr,"node_%s --> (%s , %s)\n",
	    tree->IntermediateNode[i]->refName,
	    tree->IntermediateNode[i]->branch1->refName,
	    tree->IntermediateNode[i]->branch2->refName);
  }
#endif
  if (flis != NULL) {
    fprintf(flis,"\nTree description using refNames:\n");
    DescribeTree(flis,tree,tree->root);
  }
}

#if 0
void ScanConvSequence(char *fname)
{
  DtCSRec        *pCS;
  DtSiteCSRec *sitei;
  int i;
  FILE *fp; 
  char c;

  if (fname[0] == '\0') return;
  if ((fp=fopen(fname, "r")) == NULL) {
    fprintf(stderr,"InputRootSequence:Sorry, could not open file %s\n",fname);
    exit(0);
  }

/*     fprintf(stderr, "Root Sequence File: %s (for %d bases)\n",fname,numBases); */
  for (i=0; i<numBases; i++) {
    sitei=&pCS->p[i];
    fscanf(fp,"%d",&sitei->conv);
  }
  fclose(fp);
}
#endif

/* =========================================================
 * Read reference sequences from file.
 * Supposes refNames previously input.
 *
 * Fasta format
 * =========================================================
 */
void InputRefSequences(char *fname, TTree *tree)
{
  extern TNode *inqNodeWithLabel(TTree *tree, char *Label);
  extern void WriteRefSequences();

  char buff[BUFSIZ];
  char wbuff[BUFSIZ];
  char Label[80];
  int  nAlloc = 0;
  int  same;
  int  i;
  char *pC;
  FILE *fp; 
  char c;
  TNode *node;
  int lsze = 0;

  if (fname[0] == '\0') return;
  if ((fp=fopen(fname, "r")) == NULL) return;

  while (fgets(buff, BUFSIZ, fp) != NULL) {
    if ((pC = strchr(buff,'\n')) != NULL) *pC = '\000';
    if (buff[0] == '>') {
      pC = &buff[1];
      sscanf(pC,"%s",Label);
      if ((node = inqNodeWithLabel(tree,Label)) == NULL) {
	fprintf(stderr,"Sorry: reference label %s does not correspond to a node\n",Label);
	exit(0);
      }
      
      node->refSeq = CAllocMem(numBases+1, "refSeq", "InputRefSequences", 0);
      
      node->refSeq[0] = '\0';
      lsze = 0;

    } else {
      lsze += strlen(buff);
      strcat(node->refSeq, buff);
    }
  }

  fclose(fp);


/*   WriteRefSequences("toto", tree); */
}

/* =========================================================
 * Read reference sequences from file.
 * Supposes refNames previously input.
 *
 * PAML sequential phylip like format.
 * =========================================================
 */
void InputPAMLRefSequences(char *fname, TTree *tree)
{
  extern TNode *inqNodeWithLabel(TTree *tree, char *Label);
  extern void WriteRefSequences();

  char buff[BUFSIZ];
  char wbuff[BUFSIZ];
  char Label[80];
  int  nAlloc = 0;
  int  same;
  int  i, j, debut;
  char *pC;
  FILE *fp; 
  char c;
  TNode *node;
  int lsze = 0;
  int nSeq, sSze;



  if (fname[0] == '\0') return;
  if ((fp=fopen(fname, "r")) == NULL) return;

  fprintf(stderr,"inputPAMLRefSequences: (%s)\n", fname);

  fscanf(fp,"%d%d",&nSeq,&sSze);

  fprintf(stderr,"inputPAMLRefSequences: %d sequences %d length\n",nSeq,sSze);
  fflush(stderr);

  i = 0;
  while (i<nSeq) {
    debut = 1;
    while (fgets(buff, BUFSIZ, fp) != NULL) {

      if (buff[0] == '\n') continue;
      if ((pC = strchr(buff,'\n')) != NULL) *pC = '\000';
      pC = buff;

      if (debut) {
	sscanf(buff,"%s",Label);
	
	if (DEBUG) fprintf(stderr,"inputPAMLRefSequences: sequence %s\n",Label);

	if ((node = inqNodeWithLabel(tree,Label)) == NULL) {
	  fprintf(stderr,"Sorry: reference label %s does not correspond to a node\n",Label);
	  exit(0);
	}
	
	node->refSeq = CAllocMem(numBases+1, "refSeq", "InputRefSequences", 0);
	
	node->refSeq[0] = '\0';
	lsze = 0;
	debut =0;
	pC = buff+30;
      } 

      if (DEBUG) fprintf(stderr,"%s\n",pC);

      for (j = 0; j < strlen(pC); j++) {
	if (pC[j] != ' ') node->refSeq[lsze++] = pC[j];
      }
      if (lsze == sSze) {
	debut = 1;
	node->refSeq[lsze] = '\000';
	i++;
      }
    }
    if (i < nSeq) {
      fprintf(stderr,"inputPAMLRefSequences: Sorry: Could not input %d sequences (got %d)\n",nSeq,i);
      exit(0);
    }
  }

  fclose(fp);

  if (DEBUG) fprintf(stderr,"inputPAMLRefSequences: done ...\n");

/*   WriteRefSequences("toto", tree); */
}

/* =========================================================
 * Write simulated sequences to file.
 * =========================================================
 */
void WriteSequences(FILE *fv, TTree *tree, int OutNode)    /* modified MD 97 */
{
    extern int Fasta;

    int i, j,k,l,b=0;
    char *P;
    int numtipnode;

    if (fv == NULL) return;

    if (!Fasta) fprintf(fv, " %d %d\n", numTaxa, numBases);
    numtipnode=(tree->numInteNode)-1;

/*     fprintf(stderr,"For Nodes: OutNode %d, numInteNode %d\n",OutNode,tree->numInteNode); */
    if (OutNode==1) {
    
	for (i=0; i<tree->numInteNode; i++)
	    {
		if (tree->IntermediateNode[i]->tipNo!=-1) 
		    {
			b++;
			continue;
		    }

		if ((Fasta) && (tree->IntermediateNode[i]->refName != NULL)) {
		  fprintf(fv,"> %s ",tree->IntermediateNode[i]->refName);
		  fprintf(fv,"(%s ",tree->IntermediateNode[i]->branch1->refName);
		  fprintf(fv," %s)",tree->IntermediateNode[i]->branch2->refName);
		  
		} else if (Fasta) fprintf(fv,"> ");

		if (i==0) fprintf(fv,"root");
		else {
		  fprintf(fv,"node_%d (%d) ",i-b,tree->IntermediateNode[i]->IntermediateNodeNo);
		  if (tree->IntermediateNode[i]->branch1->tipNo==-1)
		    fprintf(fv,"(%d ",tree->IntermediateNode[i]->branch1->IntermediateNodeNo);
		  else
		    fprintf(fv,"(%s ", tree->names[tree->IntermediateNode[i]->branch1->tipNo]);
		  if (tree->IntermediateNode[i]->branch2->tipNo==-1)
		    fprintf(fv,"%d)",tree->IntermediateNode[i]->branch2->IntermediateNodeNo);
		  else
		    fprintf(fv,"%s)", tree->names[tree->IntermediateNode[i]->branch2->tipNo]);
		}
		if (relaxedPhylip) {
		    if (!Fasta) fputc(' ', fv);
		    else fputc('\n', fv);
		}
		else
		    { 
			l=0;
			if (i==0) l=4;
			else {
			    if (i>=1000) l=9;
			    else {
				if (i>=100) l=8;
				else {
				    if (i>=10) l=7;
				    else l=6;
				}
			    }
			}
			
			while (l<10) 
			    {
				fputc(' ', fv);
				l++;
			    }
		    }
		P=tree->IntermediateNode[i]->sequence;
		for (j=1; j<numBases+1; j++) 
		    { 
			fprintf(fv,"%c",aminoAcids[*P]);
			P++; 
			if ((Fasta) && ((j%50)==0) && (j!=numBases)) fprintf(fv,"\n");
		    } 
		fputc('\n', fv);
	    }
    }
  

	for (i=0; i<tree->numTips; i++) {
		if (relaxedPhylip)
		    if (Fasta) {
			fprintf(fv, "> %s", tree->names[i]);
			fputc('\n', fv);
		    } else fprintf(fv, "%s ", tree->names[i]);
		else {
			j=0;
			P=tree->names[i];
			while (j<10 && *P) {
				fputc(*P, fv);
				j++;
				P++;
			}
			while (j<10) {
				fputc(' ', fv);
				j++;
			}
		}
		P=tree->tips[i]->sequence;
		for (j=1; j<numBases+1; j++) {
			fputc(aminoAcids[*P], fv);
			P++;
			if ((Fasta) && ((j%50)==0) && (j!=numBases)) fprintf(fv,"\n");
		}
		fputc('\n', fv);
	}
    if (DEBUG) fprintf(stderr, "Sequences Output done ...\n");
}


/* =========================================================
 * Write reference sequences to file.
 * =========================================================
 */
void WriteRefSequences(char *fname, TTree *tree, int OutNode)    /* modified MD 97 */
{
    extern int Fasta;

    int i, j,k,l,b=0;
    char *P;
    FILE *fv;
    int numtipnode;

    if (fname[0] == '\0') strcpy(fname,"pseqgen.refseq");
    fv=fopen(fname, "w");
    if (fv == NULL) {
      fprintf(stderr, "Sorry: could not open file %s for sequence output\n",fname);
      exit(0);
    }
    fprintf(stderr, "Sequences Output File: %s\n",fname);


    if (!Fasta) fprintf(fv, " %d %d\n", numTaxa, numBases);
    numtipnode=(tree->numInteNode)-1;

/*     fprintf(stderr,"For Nodes: OutNode %d, numInteNode %d\n",OutNode,tree->numInteNode); */
    if (OutNode==1) {
    
	for (i=0; i<tree->numInteNode; i++)
	    {
		if (tree->IntermediateNode[i]->tipNo!=-1) 
		    {
			b++;
			continue;
		    }
#if 0	
		fprintf(stderr," Node %d -> ",tree->IntermediateNode[i]->IntermediateNodeNo);
		if (tree->IntermediateNode[i]->branch1->tipNo == -1)
		  fprintf(stderr,"  %d",tree->IntermediateNode[i]->branch1->IntermediateNodeNo);
		else 
		  fprintf(stderr," %s",tree->names[tree->IntermediateNode[i]->branch1->tipNo]);
		  
		if (tree->IntermediateNode[i]->branch2->tipNo == -1)
		  fprintf(stderr,"  %d\n",tree->IntermediateNode[i]->branch2->IntermediateNodeNo);
		else 
		  fprintf(stderr," %s\n",tree->names[tree->IntermediateNode[i]->branch2->tipNo]);
		  
#endif

		if (Fasta) fprintf(fv,"> %s ",tree->IntermediateNode[i]->refName);
		
		if (i==0) fprintf(fv,"root");
		else {
		  fprintf(fv,"node_%d (%d) ",i-b,tree->IntermediateNode[i]->IntermediateNodeNo);
		  if (tree->IntermediateNode[i]->branch1->tipNo==-1)
		    fprintf(fv,"(%d ",tree->IntermediateNode[i]->branch1->IntermediateNodeNo);
		  else
		    fprintf(fv,"(%s ", tree->names[tree->IntermediateNode[i]->branch1->tipNo]);
		  if (tree->IntermediateNode[i]->branch2->tipNo==-1)
		    fprintf(fv,"%d)",tree->IntermediateNode[i]->branch2->IntermediateNodeNo);
		  else
		    fprintf(fv,"%s)", tree->names[tree->IntermediateNode[i]->branch2->tipNo]);
		}
		if (relaxedPhylip) {
		    if (!Fasta) fputc(' ', fv);
		    else fputc('\n', fv);
		}
		else
		    { 
			l=0;
			if (i==0) l=4;
			else {
			    if (i>=1000) l=9;
			    else {
				if (i>=100) l=8;
				else {
				    if (i>=10) l=7;
				    else l=6;
				}
			    }
			}
			
			while (l<10) 
			    {
				fputc(' ', fv);
				l++;
			    }
		    }
		P=tree->IntermediateNode[i]->refSeq;
		for (j=1; j<numBases+1; j++) 
		    { 
/* 			fprintf(fv,"%c",aminoAcids[*P]); */
			fprintf(fv,"%c",*P);
			P++; 
			if ((Fasta) && ((j%50)==0) && (j!=numBases)) fprintf(fv,"\n");
		    } 
		fputc('\n', fv);
	    }
    }
  

	for (i=0; i<tree->numTips; i++) {
		if (relaxedPhylip)
		    if (Fasta) {
			fprintf(fv, "> %s", tree->names[i]);
			fputc('\n', fv);
		    } else fprintf(fv, "%s ", tree->names[i]);
		else {
			j=0;
			P=tree->names[i];
			while (j<10 && *P) {
				fputc(*P, fv);
				j++;
				P++;
			}
			while (j<10) {
				fputc(' ', fv);
				j++;
			}
		}
		P=tree->tips[i]->refSeq;
		for (j=1; j<numBases+1; j++) {
/* 			fputc(aminoAcids[*P], fv); */
			fputc(*P, fv);
			P++;
			if ((Fasta) && ((j%50)==0) && (j!=numBases)) fprintf(fv,"\n");
		}
		fputc('\n', fv);
	}
    fprintf(stderr, "RefSequences Output done ...\n");
    fclose(fv);
}

/* =========================================================
 * Setup branch lengths as true number of substitutions observed.
 * (recursive)
 * =========================================================
 */
void SetBranchLenFromRefSeq(TTree *tree, TNode *node, int whichSeq, int mode)
{
  int nDiff;
  double dnDiff;
  

  /* If taxon: done ... */
  if (node->tipNo!=-1) return;
  
  /* For branch1 ... */
  if (mode == DcLENGTHSFROMOBS) {
    nDiff = GetBranchLength(node, node->branch1, whichSeq, tree->siteMode);
    dnDiff = (double) nDiff;
    fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	    node->refName,node->branch1->refName,node->length1,dnDiff);
    node->length1 = dnDiff;
    node->branch1->length0 = dnDiff;
  } else if (mode == DcLENGTHSFROMDIST) {
    dnDiff = node->length1;
    node->length1 = tree->nESites *  node->length1;
    node->branch1->length0 =  node->length1;
    fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	    node->refName,node->branch1->refName,dnDiff,node->length1);
  }
  SetBranchLenFromRefSeq(tree, node->branch1, whichSeq, mode);

  /* For branch2 ... */
  if (mode == DcLENGTHSFROMOBS) {
    nDiff = GetBranchLength(node, node->branch2, whichSeq, tree->siteMode);
    dnDiff = (double) nDiff;
    fprintf(stderr,"%s - (br2) %s : %.3lf -> %.3lf\n",node->refName,node->branch2->refName,node->length2,dnDiff);
    node->length2 = dnDiff;
    node->branch2->length0 = dnDiff;
  } else if (mode == DcLENGTHSFROMDIST) {
    dnDiff = node->length2;
    node->length2 = tree->nESites *  node->length2;
    node->branch2->length0 =  node->length2;
    fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	    node->refName,node->branch1->refName,dnDiff,node->length2);
  }
  SetBranchLenFromRefSeq(tree, node->branch2, whichSeq, mode);

  /* For branch0 if root and unrooted ... */
  if ((node == tree->root) && (!tree->rooted)) {
    if (mode == DcLENGTHSFROMOBS) {
      nDiff = GetBranchLength(node, node->branch0, whichSeq, tree->siteMode);
      dnDiff = (double) nDiff;
      fprintf(stderr,"%s - (br0) %s : %.3lf -> %.3lf\n",node->refName,node->branch0->refName,node->length0,dnDiff);
      node->length0 = dnDiff;
      node->branch0->length0 = dnDiff;
    } else if (mode == DcLENGTHSFROMDIST) {
      dnDiff = node->length0;
      node->length0 = tree->nESites *  node->length0;
      node->branch0->length0 =  node->length0;
      fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	      node->refName,node->branch1->refName,dnDiff,node->length0);
    }
    SetBranchLenFromRefSeq(tree, node->branch0, whichSeq, mode);
  } 
}


/* =================================================================
 * Generic function to perform input of reference sequences and
 * manage related information
 * =================================================================
 */
void RefSeqUse(TTree *tree, char *iRefSeqFileName, int whichTpe, int noGapMask, int obsRates, int obsBLen, double *gRefTLen)
{
  double refTLen;
  int tmpI;
  int aPos;
  
  if (iRefSeqFileName == NULL) return;
  if (iRefSeqFileName[0] == '\0') return;
	      
  if (DEBUG) fprintf(stderr,"RefSeqUse : Will input reference sequence ...\n");
  if (whichTpe == 0) InputRefSequences(iRefSeqFileName, tree);
  else  InputPAMLRefSequences(iRefSeqFileName, tree);
  if (DEBUG) fprintf(stderr,"RefSeqUse : Reference sequences input done ...\n");

  tree->refSeqp = DcREFSEQUSE;

#if 0
  /* Count number of co-subsitutions along tree for reference set */
  CS_CSTreeCount(tree, DcREFSEQ, tree->pCs->rfCsTab, tree->pCs->p,  &tree->pCs->nRfCs);
  fprintf(stderr,"count for co-subsitutions (reference set) done ...\n");
  fflush(stderr);
#endif
      
  /* Setup mask to filter evolving sites ----------------- */
  /* 1: gapped sites are locked  ------------------------- */
  if (noGapMask) SetupNoGapMask(tree, stderr, &tree->mask);
  for (aPos=0; aPos<numBases; aPos++) {
    if (tree->mask[aPos] == 0) {
      tree->siteMode[aPos] = DcCONSTANT;
    }
  }


  if (obsRates) {
    if (DEBUG) fprintf(stderr,"RefSeqUse : Will setup observed rates ...\n");
    /* 1st pass for SetupRateMask */
    tmpI = countTrueSites(tree);
    tree->nESites = (double) tmpI;
    SetRatesFromRefSeqs(tree);

    /* too conserved or too variable sites may be locked --- */
    /* Default is to keep all by assigning gMinRate to 0., gMaxRate to 1000. */
    if (DEBUG) fprintf(stderr,"RefSeqUse : Will setup rates mask ...\n");
    fflush(stderr);
    SetupRateMask(tree, stderr, &tree->mask, tree->minRate, tree->maxRate);
  
    /* 2nd pass for SetRatesFromRefSeqs with correct tree->nESites */
    tmpI = countTrueSites(tree);
    tree->nESites = (double) tmpI;
    fprintf(stderr,"Total of %d evolving sites\n",tmpI);
    SetRatesFromRefSeqs(tree);

    CS_siteMode2tpe(tree->siteMode, tree->pCs->p, tree->nSites);

    OutputRates(stderr);
  }

  /* If required, setup branch lengths as observed from the reference 
     set */
  if (obsBLen) {
    if (DEBUG) fprintf(stderr,"RefSeqUse : Will setup ref branch lengths ...\n");
    SetBranchLenFromRefSeq(tree, tree->root, DcREFSEQ, tree->distMode);
    refTLen = getTreeLength(tree, DcREFSEQ);
    *gRefTLen = refTLen;
  }
}
