/*
 * CS-PSeq-Gen addons by
 * P. Tuffery (1997-2000)
 * INSERM U436, Paris, France
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "global.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "random.h"
#include "CSFuncs.h"

extern int DEBUG;

/* ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys, met, phe, pro, ser, thr, trp, tyr, val */
static  int MHMAAClassVec[NUM_AA] = { /* According to M.H. Muchielli */
  1,   5,   3,   0,   6,   2,   0,   3,   4,    1,   1,   5,   1,   1,   2,   3,   3,   2,   2,   1};

static  int HBAAClassVec[NUM_AA] = { /* According to M.H. Muchielli */
   0,   1,   1,   1,   0,   1,   1,   0,   1,    0,   0,   1,   0,   0,   0, 
   1,   1,   0,   0,   0};
  
static  int dummy1MHMAAClassVec[NUM_AA] = { /* Dummy 7 classes */
  1,   1,   1,   1,   1,   1,   0,   0,   2,    2,   2,   2,   3,   3,   3,   3,   4,   5,   5,   6};
  
static  int dummy2MHMAAClassVec[NUM_AA] = { /* Dummy 7 classes */
  1,   6,   1,   0,   1,   3,   1,   0,   2,    3,   2,   4,   3,   1,   3,   2,   5,   2,   5,   1};
  
static  int PolarAAClassVec[NUM_AA] = { /* Classic polar classes */
  0,   1,   3,   2,   0,   3,   2,   0,   0,    0,   0,   1,   0,   0,   0,   3,   3,   0,   0,   0};
      
static  int AAClassVec[NUM_AA] = { /* All different */
  0,   1,   2,   3,   4,   5,   6,   7,   8,    9,  10,  11,  12,  13,  14,  15,  16,  17,  18,  19};

static int *pAAClass = AAClassVec;

static double matrix  [DcMAXNUMAA];
static double matrixnc  [DcMAXNUMAA];
static double matrix2 [DcMAXNUMAA];

int commut[DcMAXNUMAA][DcMAXNUMAA];
static int smcommut[DcMAXNUMAA][DcMAXNUMAA];
int rfcommut[DcMAXNUMAA][DcMAXNUMAA];
static int iscommut[DcMAXNUMAA][DcMAXNUMAA];

int gNCmt, gNRfCmt;

void setAAClassFilter()
{
  pAAClass = AAClassVec;
}

void setMHMAAClassFilter()
{
  pAAClass = MHMAAClassVec;
}

void setHBAAClassFilter()
{
  pAAClass = HBAAClassVec;
}

void setDummy1MHMAAClassFilter()
{
  pAAClass = dummy1MHMAAClassVec;
}

void setDummy2MHMAAClassFilter()
{
  pAAClass = dummy2MHMAAClassVec;
}

void setPolarAAClassFilter()
{
  pAAClass = PolarAAClassVec;
}

void pAAClassInfo()
{
  if (pAAClass == AAClassVec) fprintf(stderr,"AA classes           : standart 20 states\n");
  else if (pAAClass == MHMAAClassVec) fprintf(stderr,"AA classes           : MHM 7 states\n");
  else if (pAAClass == dummy1MHMAAClassVec) fprintf(stderr,"AA classes           : Dummy1 7 states\n");
  else if (pAAClass == dummy2MHMAAClassVec) fprintf(stderr,"AA classes           : Dummy2 7 states\n");
  else if (pAAClass == PolarAAClassVec) fprintf(stderr,"AA classes           : Polar 4 states\n");
}

int WichAminoAcidAAClass(char c)
{
  int aa;

  aa = WhichAminoAcid(c);
  if (aa == 255) return -1;
  if (aa == NUM_AA) return -2;
  return pAAClass[aa]; 
}

int WichAAClass(int aa)
{

  if (aa == 255) return -1;
  if (aa == NUM_AA) return -2;
  return pAAClass[aa]; 
}

/* =======================================================
 * Compute exp(entropies) for tip sequences.
 * =======================================================
 */
#if 0
static double cumH[DcMAXNUMAA][DcAATYPES];
static double P[DcMAXNUMAA];
static double lP1[DcMAXNUMAA];
static double P1[DcMAXNUMAA];
static double P2[DcMAXNUMAA];

static double H[DcMAXNUMAA];
static double GH[DcMAXNUMAA];
static double GH1[DcMAXNUMAA];
static double GH2[DcMAXNUMAA];

void GetEntropies(TTree *tree)    /* PT 98 */
{
  int i,j;
  float tmpFloat;
  double tmpDouble;
  double sumH;
  double lH;

  tmpFloat = (float) tree->numTips;
  tmpDouble = (double) tmpFloat;

  memset((char *) cumH, 0, DcAATYPES * DcMAXNUMAA * sizeof(double));

/*   printf("numBases is %d tips is %d (%.0lf)\n",numBases,tree->numTips, tmpDouble); */
  for (i=0; i<numBases; i++) {
    for (j=0; j<tree->numTips; j++) {
/*       if (i == 1) printf("Tip %d site %d : %d (%c)\n",i,j,tree->tips[j]->sequence[i],aminoAcids[tree->tips[j]->sequence[i]]); */
      cumH[i][tree->tips[j]->sequence[i]] += 1.;
    }
  }

  for (i=0; i<numBases; i++) {
    for (j=0; j<DcAATYPES; j++) {
      cumH[i][j] /= tmpDouble;
/*       if (i == 1) printf("Tip %d tpe %d : cum  %lf\n",i,j,cumH[i][j]); */
    }
  }

  for (i=0; i<numBases; i++) {
    H[i] = 0.;
    for (j=0; j<DcAATYPES; j++) {
      if (cumH[i][j] > 0.00000000000000001)
	H[i] += - cumH[i][j] * log(cumH[i][j]);
    }

/*     printf("Site %d H is %lf\n",i,H[i]); */
    H[i] = exp(H[i]) - 1.;
/*     printf("Site %d exp(H)-1 is %lf\n",i,H[i]); */
  }

  tmpFloat = (float) numBases;
  tmpDouble = (double) tmpFloat;
  sumH = 0.;
  for (i=0; i<numBases; i++) {
      sumH += H[i];
  }
  sumH /= tmpDouble;

  for (i=0; i<numBases; i++) {
    GH[i]  += H[i];
    lH = H[i] / sumH; 
    GH1[i] += lH;
    GH2[i] += lH * lH;
  }
}

void initEntropy()
{
  int i;
  for (i=0; i<numBases; i++) {
    GH[i] = 0.;
    GH1[i] = 0.;
    GH2[i] = 0.;
  }
}

void OutputEntropies(FILE *f)
{
  int i;
  float tmpFloat;
  extern int numDatasets;
  double sumH = 0., dNumBases, dNumSets;

  tmpFloat  = (float) numBases;
  dNumBases = (double) tmpFloat;

  tmpFloat  = (float) numDatasets;
  dNumSets = (double) tmpFloat;
  for (i=0; i<numBases; i++) {
    sumH += GH[i];
  }
  printf("Mean relative variation of equivalent number of class per site. (Mean is %lf)\n",
	 sumH / (dNumBases * dNumSets));
  sumH /= (dNumBases);
  for (i=0; i<numBases; i++) {
/*     GH[i] /= dNumSets; */
    fprintf(f, "%6.3lf ", GH[i] / sumH);
    if (!((i+1)%10)) fprintf(f,"\n");
  }
  fprintf(f,"\n");

  /* variance is 1/(n-1)[SumX2 - ((SumX)2/n)] */
  printf("STdev of  relative variation of equivalent number of class per site.\n");
  for (i=0; i<numBases; i++) {
    fprintf(f, "%6.3lf ",(GH2[i] - (GH1[i]*GH1[i] / dNumSets)) / dNumSets);
    if (!((i+1)%10)) fprintf(f,"\n");
  }
  fprintf(f,"\n");

}
#endif

void initTreeCount(TTree *tree)
{
  if (tree->sumS == NULL)
    tree->sumS  = malloc(sizeof(double) * (tree->nSites + 1));
  if (tree->sumS2 == NULL)
    tree->sumS2 = malloc(sizeof(double) * (tree->nSites + 1));
  if (tree->aDVec == NULL)
    tree->aDVec = malloc(sizeof(double) * (tree->nSites + 1));

  memset(tree->sumS, 0,sizeof(double) * (tree->nSites + 1));
  memset(tree->sumS2,0,sizeof(double) * (tree->nSites + 1));
  memset(tree->aDVec,0,sizeof(double) * (tree->nSites + 1));
}

/* =========================================================
 * return number of substitutions along one given branch.
 * =========================================================
 */
int GetBranchLength(TNode *anc, TNode *des, int whichSeq, int *siteMode)
{
  int i;
  int nDiff = 0;

  if (whichSeq == DcREFSEQ) {
    for (i=0; i<numBases; i++) {
      if (siteMode[i] == DcCONSTANT) continue;
      if (WhichAminoAcid(des->refSeq[i]) != WhichAminoAcid(anc->refSeq[i])) {
	++nDiff;
      }
    }
  } else {
    for (i=0; i<numBases; i++) {
      if (siteMode[i] == DcCONSTANT) continue;
      if (des->sequence[i] != anc->sequence[i]) {
	++nDiff;
      }
    }
  }
  return nDiff;
}

/* =========================================================
 * Setup branch lengths as true number of substitutions observed.
 * (recursive)
 * =========================================================
 */
void SetLengthsAsTrueLengths(TTree *tree, TNode *node, int whichSeq, int mode)
{
  int nDiff;
  double dnDiff;
  

  /* If taxon: done ... */
  if (node->tipNo!=-1) return;
  
  /* For branch1 ... */
  if (mode == DcLENGTHSFROMOBS) {
    nDiff = GetBranchLength(node, node->branch1, whichSeq, tree->siteMode);
    dnDiff = (double) nDiff;
    fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	    node->refName,node->branch1->refName,node->length1,dnDiff);
    node->length1 = dnDiff;
    node->branch1->length0 = dnDiff;
  } else if (mode == DcLENGTHSFROMDIST) {
    dnDiff = node->length1;
    node->length1 = tree->nESites *  node->length1;
    node->branch1->length0 =  node->length1;
    fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	    node->refName,node->branch1->refName,dnDiff,node->length1);
  }
  SetLengthsAsTrueLengths(tree, node->branch1, whichSeq, mode);

  /* For branch2 ... */
  if (mode == DcLENGTHSFROMOBS) {
    nDiff = GetBranchLength(node, node->branch2, whichSeq, tree->siteMode);
    dnDiff = (double) nDiff;
    fprintf(stderr,"%s - (br2) %s : %.3lf -> %.3lf\n",node->refName,node->branch2->refName,node->length2,dnDiff);
    node->length2 = dnDiff;
    node->branch2->length0 = dnDiff;
  } else if (mode == DcLENGTHSFROMDIST) {
    dnDiff = node->length2;
    node->length2 = tree->nESites *  node->length2;
    node->branch2->length0 =  node->length2;
    fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	    node->refName,node->branch1->refName,dnDiff,node->length2);
  }
  SetLengthsAsTrueLengths(tree, node->branch2, whichSeq, mode);

  /* For branch0 if root and unrooted ... */
  if ((node == tree->root) && (!tree->rooted)) {
    if (mode == DcLENGTHSFROMOBS) {
      nDiff = GetBranchLength(node, node->branch0, whichSeq, tree->siteMode);
      dnDiff = (double) nDiff;
      fprintf(stderr,"%s - (br0) %s : %.3lf -> %.3lf\n",node->refName,node->branch0->refName,node->length0,dnDiff);
      node->length0 = dnDiff;
      node->branch0->length0 = dnDiff;
    } else if (mode == DcLENGTHSFROMDIST) {
      dnDiff = node->length0;
      node->length0 = tree->nESites *  node->length0;
      node->branch0->length0 =  node->length0;
      fprintf(stderr,"%s - (br1) %s : %.3lf -> %.3lf\n",
	      node->refName,node->branch1->refName,dnDiff,node->length0);
    }
    SetLengthsAsTrueLengths(tree, node->branch0, whichSeq, mode);
  } 
}

/* =========================================================
 * Count tree length in terms of substitutions along tree.
 * (on refseq (whichSeq = DcREFSEQ) or sequences (whichSeq = DcSIMSEQ))
 * =========================================================
 */
double getTreeLength(TTree *tree, int whichSeq)
{
  int i;
  double sumP = 0.;

  CS_SSTreeCount(tree, whichSeq, tree->aDVec,tree->nSites);
  for (i=0; i<numBases; i++) {
    if (tree->siteMode[i] == DcCONSTANT) continue;
    sumP += tree->aDVec[i];
  }

  return sumP;
}


/* =========================================================
 * Count tree length in terms of substitutions along tree.
 * (on refseq (whichSeq = DcREFSEQ) or sequences (whichSeq = DcSIMSEQ))
 * =========================================================
 */
void SumTreeDist(TTree *tree, TNode *node, double *aSum)
{

  *aSum += node->length1;
  if ( node->tipNo !=-1) return;

  SumTreeDist(tree, node->branch1, aSum);
  *aSum += node->length2;
  SumTreeDist(tree, node->branch2, aSum);
  if (!tree->rooted && (node == tree->root)) {
    *aSum += node->length0;
    SumTreeDist(tree, node->branch0, aSum);
  }

}


/* ====================================================
 * store number of mutations per sites over simulations
 * P stores exact count
 * P1 stores mean(X)
 * P2 stores mean(X)*mean(X)
 * ====================================================
 */
void storeTreeLengths(TTree *tree, int whichSeq)
{
  int i;
  double sumP = 0.;
  float tmpFloat;
  double tmpDouble;

  CS_SSTreeCount(tree, whichSeq, tree->aDVec,tree->nSites);
  for (i=0; i<tree->nSites; i++) {
    if (tree->siteMode[i] == DcCONSTANT) continue;
    sumP += tree->aDVec[i];
  }
  
  sumP /= tree->nESites;

  for (i=0; i<tree->nSites; i++) {
    if (tree->siteMode[i] == DcCONSTANT) continue;
    tree->sumS[i] += tree->aDVec[i];

    tree->aDVec[i] /= sumP;
    tree->sumS2[i] += tree->aDVec[i] * tree->aDVec[i];
  }
}





#if 0

/* ====================================================
 * To write number of mutations observed.
 * ====================================================
 */
void OutputTreeCount(FILE *f)
{
  int i;
  float tmpFloat;
  extern int numDatasets;
  double sumP = 0., dNumBases, dNumSets, sumR = 0.;
  int nSites = 0;

  tmpFloat  = (float) numDatasets;
  dNumSets = (double) tmpFloat;

  for (i=0; i<numBases; i++) {
    if (CM[i].MutAllowed == 'n') continue;
    ++nSites;
    sumP += P[i];
  }
  fprintf(f,"Mean number of mutations along tree per site (%d) is %lf\n",
/*  	 sumP / (dNumBases * dNumSets)); */
	 nSites, sumP / (gNTrueSites * dNumSets));

#if 0 /* OLD */
  tmpFloat  = (float) numBases;
  dNumBases = (double) tmpFloat;

  sumP /= (dNumBases);
#endif
  sumP /= gNTrueSites; 

  for (i=0; i<numBases; i++) {
    fprintf(f, "%6.3lf ", P[i] / sumP);
    sumR += P[i] / sumP;
    if (!((i+1)%10)) fprintf(f,"\n");
  }
  fprintf(f,"\n");
  fprintf(f,"Sum of rates is %lf (%d sites)\n",sumR, nSites);

  /* variance is 1/(n-1)[SumX2 - ((SumX)2/n)] */
  fprintf(f,"STdev of mean number of mutations along tree per site.\n");
  for (i=0; i<numBases; i++) {
    fprintf(f, "%6.3lf ",(P2[i] - (P1[i]*P1[i] / dNumSets)) / dNumSets);
    if (!((i+1)%10)) fprintf(f,"\n");
  }
  fprintf(f,"\n");
}


/* =========================================================
 * copy commut into rfcommut
 * =========================================================
 */
void commutAsRfCommut()
{
  int i,j;

  for (i=0; i<numBases; i++) {
    for (j=0; j<numBases; j++) {
      rfcommut[i][j]   = commut[i][j];
    }
  }
}
/* =========================================================
 * Initialization of commut arrays
 * =========================================================
 */
void initLclCommut()
{
  int i,j;

  gNCmt = 0;
  for (i=0; i<numBases; i++) {
    for (j=0; j<numBases; j++) {
      commut[i][j]   = 0;
    }
  }
}


/* =========================================================
 * Initialization of commut arrays
 * =========================================================
 */
void initRfCommut()
{
  int i,j;

  gNCmt = 0;
  for (i=0; i<numBases; i++) {
    for (j=0; j<numBases; j++) {
      rfcommut[i][j]   = 0;
    }
  }
}


#endif


