/*  Header file to define tree and node structures     */

/*  (c) Copyright 1996, Andrew Rambaut                 */
/*  Department of Zoology, University of Oxford        */

#ifndef _TREE_H_
#define _TREE_H_

#define MAX_NAME_LEN 64
#define MAX_TIPS 1000

#define DcSIMPLE 1
#define DcCOR    2
#define DcBOT    3

#define DcCORRNOTALLOW 0
#define DcCORRALLOW    1

#define DcSIMSEQ 0
#define DcREFSEQ 1

#define DcAATYPES  20
#define DcMAXNUMAA 400
#define DcMAXCOMUT 50
#define DcMAXGLOBCOMUT 50000
#define DcLENGTHSFROMTREEFILE 0
#define DcLENGTHSFROMOBS      1
#define DcLENGTHSFROMDIST     2

#define DcRANDOMROOT           0  /* Root sequence is generated      */
#define DcUSERROOT             1  /* Root sequence is read from disk */
#define DcUSEREQUILIBRATEDROOT 2  /* Root sequence is equilibrated   */

#define DcREFSEQNOUSE         0   /* No reference sequences          */
#define DcREFSEQUSE           1   /* reference sequences             */

#include "CSTypes.h"

typedef struct TNode TNode;
struct TNode {
  TNode *branch0, *branch1, *branch2, *next;
  double length0, length1, length2, param;
  int tipNo;

  int IntermediateNodeNo;   /* M D 1997 */
  int   genMode;            /* Generated by simple, BOT or COR, P. TUFFERY, 1998 */
  char *sequence;
  char *refName;            /* P T 1998: reference Id (for interfacing with PAUP) */
  char *refSeq;             /* P T 1998: reference sequence */

};


typedef struct TTree TTree;
struct TTree {
  int rooted, lengths;
  TNode *root, *nodeList;
  int numTips, numNodes;
  int numInteNode;                    /* M D 1997 */
  int *mask;                          /* P T 1999 */
  double totalLength;
  char *names[MAX_TIPS];
  TNode *tips[MAX_TIPS];
  TNode *IntermediateNode[MAX_TIPS];  /* M D 1997 */
  double refTLen;                     /* I */

  /* A small addendum focused on sites and not on tree itself */
  /* But helpful to propagate pointers in functions           */
  int     nSites;                     /* I P. Tuffery 2000 (== numBases) */
  double  nESites;                    /* I P. Tuffery 2000 */
  int     rootMode;                   /* I P. Tuffery 2000 */
  int     refSeqp;                    /* I P. Tuffery 2000 */
  int    *siteMode;                   /* I P. Tuffery 2000 */
  double *rates;                      /* P. Tuffery 2001 */

  int     distMode;                   /* I */
  double  minRate;
  double  maxRate;

  double *sumS;                       /* I */
  double *sumS2;                      /* I */
  double *aDVec;                      /* I */

  DtCSRec *pCs;                       /* I P. Tuffery 2001 (-> CSTypes.h) */
};

#endif /* _TREE_H_ */
